import typing
from typing import Iterable
from typedpy.structures import REQUIRED_FIELDS, StructMeta, Structure, _init_class_dict

T = typing.TypeVar("T")


class PartialMeta(type):
    def __getitem__(
        cls: T, clazz: typing.Union[StructMeta, typing.Tuple[StructMeta, str]]
    ) -> T:
        if not isinstance(clazz, StructMeta):
            if not isinstance(clazz, tuple) or (
                isinstance(clazz, tuple)
                and (
                    len(clazz) != 2
                    or not isinstance(clazz[0], StructMeta)
                    or not isinstance(clazz[1], str)
                )
            ):
                raise TypeError(
                    "Partial must have a Structure class as a parameter, and an optional name for the class"
                )
        clazz, classname = (
            clazz if isinstance(clazz, tuple) else (clazz, f"Partial{clazz.__name__}")
        )

        cls_dict = _init_class_dict(clazz)
        for k, v in clazz.get_all_fields_by_name().items():
            cls_dict[k] = v

        cls_dict[REQUIRED_FIELDS] = []

        new_class = type(classname, (Structure,), cls_dict)

        return new_class


class Partial(metaclass=PartialMeta):
    """
    Define a new Structure class with all the fields of the given class, but all of them are optional.
    For Example:

     .. code-block:: python

         class Foo(ImmutableStructure):
            i: int
            d: dict[str, int] = dict
            s: str
            a: set

         class Bar(Partial[Foo]):
            x: str

    "Bar" has all the fields of Foo as optional, and in addition "x" as required. Note that Bar does not extend Foo, but
    it is a Structure class. It does not copy the serialization mappers. It does copy other attributes, such as _ignore_none,
    but Bar can override any of them.

    Another valid usage:

    .. code-block:: python

        Bar = Partial[Foo]

        bar = Bar(i=5)

        # or if you want to have a consistent class name for troubleshooting:
        Bar = Partial[Foo, "Bar"]

    """

    _required = []


class AllFieldsRequiredMeta(type):
    def __getitem__(
        cls, clazz: typing.Union[StructMeta, typing.Tuple[StructMeta, str]]
    ) -> StructMeta:
        if not isinstance(clazz, StructMeta):
            if not isinstance(clazz, tuple) or (
                isinstance(clazz, tuple)
                and (
                    len(clazz) != 2
                    or not isinstance(clazz[0], StructMeta)
                    or not isinstance(clazz[1], str)
                )
            ):
                raise TypeError(
                    "Partial must have a Structure class as a parameter, and an optional name for the class"
                )
        clazz, classname = (
            clazz
            if isinstance(clazz, tuple)
            else (clazz, f"AllFieldsRequired{clazz.__name__}")
        )

        cls_dict = _init_class_dict(clazz)
        cls_dict[REQUIRED_FIELDS] = []
        for k, v in clazz.get_all_fields_by_name().items():
            cls_dict[k] = v
            if getattr(v, "_default") is None:
                cls_dict[REQUIRED_FIELDS].append(k)

        newclass = type(classname, (Structure,), cls_dict)

        return newclass


class AllFieldsRequired(metaclass=AllFieldsRequiredMeta):
    """
    Define a new Structure class with all the fields of the given class, and all are required (even if they were
    not required in the reference class). The exception is fields that were defined with explicit default value.

    For Example:

     .. code-block:: python

         class Foo(ImmutableStructure):
            i: int
            d: dict[str, int] = dict
            s: str
            a: set

         class Bar(AllFieldsRequired[Foo]):
            x: str

    "Bar" has all the fields of Foo as required with the exception of "d" (since it has a default value), and in
     addition "x" as required.
     Note that Bar does not extend Foo, but it is a Structure class. It does not copy the serialization mappers.
     It does copy other attributes, such as _ignore_none,
     but Bar can override any of them.

    Another valid usage:

    .. code-block:: python

        Bar = AllFieldsRequired[Foo]

        bar = Bar(i=5)

        # or if you want to have a consistent class name for troubleshooting:
        Bar = AllFieldsRequired[Foo, "Bar"]

    """

    _required = []


class ExtendMeta(type):
    def __getitem__(
        cls, clazz: typing.Union[StructMeta, typing.Tuple[StructMeta, str]]
    ):
        if not isinstance(clazz, StructMeta):
            if not isinstance(clazz, tuple) or (
                isinstance(clazz, tuple)
                and (
                    len(clazz) != 2
                    or not isinstance(clazz[0], StructMeta)
                    or not isinstance(clazz[1], str)
                )
            ):
                raise TypeError(
                    "Extend must have a Structure class as a parameter, and an optional name for the class"
                )
        clazz, classname = (
            clazz if isinstance(clazz, tuple) else (clazz, f"Extend{clazz.__name__}")
        )
        cls_dict = _init_class_dict(clazz)
        cls_dict[REQUIRED_FIELDS] = getattr(clazz, REQUIRED_FIELDS, [])
        for k, v in clazz.get_all_fields_by_name().items():
            cls_dict[k] = v

        newclass = type(classname, (Structure,), cls_dict)

        return newclass


class Extend(metaclass=ExtendMeta):
    """
    Define a new Structure class with all the fields of the given class
    For Example:

     .. code-block:: python

         class Foo(ImmutableStructure):
            i: int
            d: dict[str, int] = dict
            s: str
            a: set

         class Bar(Extend[Foo]):
            x: str

    "Bar" has all the fields of Foo, and in addition "x". So the required fields in this case are i, s, a, x
    (since d has a default value).
    Note that Bar does not extend Foo, but it is a Structure class.
    It does not copy the serialization mappers. It does copy other attributes, such as _ignore_none,
    but Bar can override any of them.

    Another valid usage:

    .. code-block:: python

        Bar = Extend[Foo]

        bar = Bar(i=5)

        # or if you want to have a consistent class name for troubleshooting:
        Bar = Partial[Foo, "Bar"]

    """

    pass


class OmitMeta(type):
    def __getitem__(
        cls,
        params: typing.Union[
            typing.Tuple[StructMeta, Iterable[str], str],
            typing.Tuple[StructMeta, Iterable[str]],
        ],
    ):
        if (
            not isinstance(params, tuple)
            or not 1 < len(params) < 4
            or not isinstance(params[0], StructMeta)
        ):
            raise TypeError(
                "Omit accepts the source class name, a list of fields, and an optional name for the new class it creates"
            )
        clazz, fields, *new_name_maybe = params
        class_name = new_name_maybe[0] if new_name_maybe else f"Omit{clazz.__name__}"
        return clazz.omit(*fields, class_name=class_name)


class Omit(metaclass=OmitMeta):
    """
    Define a new Structure class with all the fields of the given class, except for the omitted ones.
     For Example:

     .. code-block:: python

         class Foo(ImmutableStructure):
            i: int
            d: dict[str, int] = dict
            s: set
            a: str
            b: Integer

         class Bar(Omit[Foo, ("a", "b")]):
            x: int

    "Bar" has the fields: i, d, s, x. Note that Bar does not extend Foo, but it is a Structure class.
    It does not copy the serialization mappers. It does copy other attributes, such as _ignore_none,
    but Bar can override any of them.

    Another valid usage:

    .. code-block:: python

        Bar = Omit[Foo, ("a", "b", "i", "s"), class_name="Bar"]
        bar = Bar(d={"a": 5})

    """

    pass


class PickMeta(type):
    def __getitem__(
        cls,
        params: typing.Union[
            typing.Tuple[StructMeta, Iterable[str], str],
            typing.Tuple[StructMeta, Iterable[str]],
        ],
    ):
        if (
            not isinstance(params, tuple)
            or not 1 < len(params) < 4
            or not isinstance(params[0], StructMeta)
        ):
            raise TypeError(
                "Pick accepts the source class name, a list of fields, and an optional name for the new class it "
                "creates "
            )
        clazz, fields, *new_name_maybe = params
        class_name = new_name_maybe[0] if new_name_maybe else f"Pick{clazz.__name__}"
        return clazz.pick(*fields, class_name=class_name)


class Pick(metaclass=PickMeta):
    """
    Define a new Structure class with that picks specific fields from a predefined class.
     For Example:

     .. code-block:: python

         class Foo(ImmutableStructure):
            i: int
            d: dict[str, int] = dict
            s: set
            a: str
            b: Integer

         class Bar(Pick[Foo, ("a", "b")]):
            x: int

    "Bar" has the fields: a, b, x. Note that Bar does not extend Foo, but it is a Structure class.
     It does not copy the serialization mappers. It does copy other attributes, such as _ignore_none,
    but Bar can override any of them.

    Another valid usage:

    .. code-block:: python

        Bar = Pick[Foo, ("d"), class_name="Bar"]
        bar = Bar(d={"a": 5})

    """

    pass



import typing
from collections import deque
from pathlib import Path

from .commons import python_ver_36, python_ver_atleast_39, python_ver_atleast_than_37


def type_is_generic(v):
    """
    Return whether or not the given type is a "standard generic",
    such as list[int], List[int], dict[int, str]
    :param v: the type
    :return: True if it is a generic, False otherwise
    """

    class Foo:
        pass

    origin = getattr(v, "__origin__", None)
    typing_base = getattr(typing, "_TypingBase", Foo)
    generic_alias = getattr(typing, "_GenericAlias", Foo)
    special_generic_alias = getattr(typing, "_SpecialGenericAlias", Foo)
    return (
        (python_ver_36 and isinstance(v, typing_base))
        or (
            python_ver_atleast_than_37
            and isinstance(v, (generic_alias, special_generic_alias))
        )
        or (
            python_ver_atleast_39
            and origin in {list, dict, tuple, set, frozenset, deque, typing.Union}
        )
    )


def get_list_type(the_type):
    """
    If the_type is a generic list type (e.g. list[int]), returns a Tuple
    of (list, <the internal type>). otherwise throws a TypeError
    :param the_type: a type that is supposed to be a "generic list", such as list[str]
    :return: a tuple of (list, <the internal type>)
    """
    if type_is_generic(the_type):
        origin = getattr(the_type, "__origin__", None)
        if origin == list:
            args = getattr(the_type, "__args__", [None])
            if len(args) == 1:
                return list, args[0]
    raise TypeError("not a list of a single type")


def maybe(func, default_val=None):
    try:
        return func()
    except AttributeError:
        return default_val


def get_abs_path_from_here(relative_path: str, calling_file_name: str) -> Path:
    """
    return absolute path to file, based on path from the calling location
    :param calling_file_name: optional name of the calling script that will serce as reference
    :param relative_path: relative path from the loading source file
    :return: absolute path with resolved symlinks
    """
    calling_source_file = calling_file_name
    full_path: Path = Path(calling_source_file).parent / relative_path
    return full_path.resolve()



import fnmatch
import logging
import subprocess
from os import walk
from pathlib import Path
import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("src_root_dir", help="source root directory")
    parser.add_argument("directory", help="directory to process")
    parser.add_argument(
        "-s",
        "--stubs-dir",
        type=str,
        default=".stubs",
        help="source directory of stubs. Default is .stubs",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        help="exclude patterns in the form path1:path2:path3",
    )
    args = parser.parse_args()
    src_root_abs_path = str(Path(args.src_root_dir).resolve())
    input_dir = str(Path(args.directory).resolve())
    stub_dir_abs_path = str(Path(src_root_abs_path) / Path(args.stubs_dir))
    exclude = args.exclude.split(":") if args.exclude else []

    for (dirpath, _, filenames) in walk(input_dir):
        if dirpath.startswith("__"):
            continue
        python_scripts = [name for name in filenames if fnmatch.fnmatch(name, "*.py")]
        for name in python_scripts:
            exclude_it = False
            abs_file_path = str(Path(dirpath) / name)
            for x in exclude:
                if fnmatch.fnmatch(abs_file_path, x):
                    exclude_it = True
                    continue
            if exclude_it:
                continue
            print(f"processing {abs_file_path}")
            try:
                with subprocess.Popen(
                    [
                        "create-stub",
                        "-s",
                        stub_dir_abs_path,
                        src_root_abs_path,
                        abs_file_path,
                    ],
                    stdout=subprocess.PIPE,
                ) as proc:
                    res = proc.stdout.read().decode("UTF-8")
                    if res:
                        print(res)
                print(f"done with {name}")
            except Exception as e:
                logging.exception(e)


if __name__ == "__main__":
    main()



import builtins
import enum
import importlib.util
import inspect
import logging
import os
import sys
import typing
from os.path import relpath
from pathlib import Path

from .commons import doublewrap_val, nested
from .fields import FunctionCall
from .serialization_wrappers import Deserializer, Serializer
from .structures import Field, ImmutableStructure, Structure
from .type_info_getter import get_all_type_info, get_type_info, is_typeddict
from .types_ast import (
    extract_attributes_from_init,
    functions_to_str,
    get_imports,
    get_models,
    models_to_src,
)
from .utility import type_is_generic

builtins_types = [
    getattr(builtins, k)
    for k in dir(builtins)
    if isinstance(getattr(builtins, k), type)
]

module = getattr(inspect, "__class__")

INDENT = " " * 4

AUTOGEN_NOTE = [
    "",
    "#### This stub was autogenerated by Typedpy",
    "###########################################",
    "",
]


_private_to_public_pkg = {"werkzeug.localxxx": "flask"}


def _get_package(v, attrs):
    pkg_name = attrs.get("__package__") or "%^$%^$%^#"
    if v.startswith(pkg_name):
        return v[len(pkg_name) :]
    return _private_to_public_pkg.get(v, v)


def _as_something(k, attrs):
    return f" as {k}" if attrs.get("__file__", "").endswith("__init__.py") else ""


def _get_struct_classes(attrs, only_calling_module=True):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and issubclass(v, Structure)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
            and v not in {Deserializer, Serializer, ImmutableStructure, FunctionCall}
        )
    }


def _get_imported_classes(attrs):
    def _valid_module(v):
        return hasattr(v, "__module__") and attrs["__name__"] != v.__module__

    res = []
    for k, v in attrs.items():
        if (
            not k.startswith("_")
            and (isinstance(v, module) or _valid_module(v))
            and not _is_internal_sqlalchemy(v)
        ):
            if isinstance(v, module):
                if v.__name__ != k:
                    parts = v.__name__.split(".")
                    if len(parts) > 1:
                        first_parts = parts[:-1]
                        last_part = parts[-1]
                        res.append(
                            (
                                k,
                                f"from {'.'.join(first_parts)} import {last_part} as {k}",
                            )
                        )
                    elif nested(lambda: getattr(v.os, k)) == v:
                        res.append((k, f"from os import {k} as {k}"))
                else:
                    res.append((k, f"import {k}"))
            else:
                pkg = _get_package(v.__module__, attrs)

                name_in_pkg = getattr(v, "__name__", k)
                name_to_import = name_in_pkg if name_in_pkg and name_in_pkg != k else k
                as_import = (
                    _as_something(k, attrs) if name_to_import == k else f" as {k}"
                )
                import_stmt = f"from {pkg} import {name_to_import}{as_import}"
                if pkg == "__future__":
                    res = [(k, import_stmt)] + res
                else:
                    res.append((k, import_stmt))
    return res


def _get_ordered_args(unordered_args: dict):
    optional_args = {k: v for k, v in unordered_args.items() if v.endswith("= None")}
    mandatory_args = {k: v for k, v in unordered_args.items() if k not in optional_args}
    return {**mandatory_args, **optional_args}


def _get_mapped_extra_imports(additional_imports) -> dict:
    mapped = {}
    for c in additional_imports:
        try:
            name = get_type_info(c, {}, set())
            if inspect.isclass(c) and issubclass(c, Structure):
                module_name = c.__module__
            else:
                module_name = (
                    c.get_type.__module__
                    if isinstance(c, Field)
                    else c.__module__
                    if name != "Any"
                    else None
                )
            if module_name:
                mapped[name] = module_name
        except Exception as e:
            logging.exception(e)
    return mapped


def _is_sqlalchemy(attr):
    module_name = getattr(attr, "__module__", "")
    return module_name and (
        module_name.startswith("sqlalchemy.orm")
        or module_name.startswith("sqlalchemy.sql")
    )


def _is_internal_sqlalchemy(attr):
    module_name = getattr(attr, "__module__", "")
    return module_name in {"sqlalchemy.orm.decl_api", "sqlalchemy.sql.schema"}


def _try_extract_column_type(attr):
    if attr.__class__.__name__ == "InstrumentedAttribute":
        return next(iter(attr.expression.base_columns)).type.python_type


def _skip_sqlalchemy_attribute(attribute):
    return attribute.startswith("_") or attribute in {"registry", "metadata"}


def _get_method_and_attr_list(cls, members):
    all_fields = cls.get_all_fields_by_name() if issubclass(cls, Structure) else {}
    ignored_methods = (
        dir(Structure)
        if issubclass(cls, Structure)
        else dir(Field)
        if issubclass(cls, Structure)
        else dir(enum)
        if issubclass(cls, enum.Enum)
        else {}
    )
    private_prefix = "_" if issubclass(cls, enum.Enum) else "__"
    method_list = []
    attrs = []
    cls_dict = cls.__dict__
    for attribute in members:
        if attribute.startswith(private_prefix) and attribute not in {
            "__init__",
            "__call__",
        }:
            continue
        attr = (
            cls_dict.get(attribute)
            if attribute in cls_dict
            else getattr(cls, attribute, None)
        )
        if _is_sqlalchemy(attr):
            if not _skip_sqlalchemy_attribute(attribute):
                members[attribute] = _try_extract_column_type(attr)
                attrs.append(attribute)
            continue
        is_func = not inspect.isclass(attr) and (
            callable(attr) or isinstance(attr, (property, classmethod, staticmethod))
        )
        if all(
            [
                attr is not None,
                not inspect.isclass(attr),
                not is_func,
                not (is_typeddict(cls) or isinstance(attr, cls)),
                not (issubclass(cls, Structure) and attribute.startswith("_")),
            ]
        ):
            attrs.append(attribute)
            continue

        if (
            is_func
            and attribute not in all_fields
            and (
                attribute not in ignored_methods
                or (issubclass(cls, Structure) and attribute == "__init__")
            )
        ):
            method_list.append(attribute)

    if (
        not issubclass(cls, Structure)
        and not issubclass(cls, enum.Enum)
        and "__init__" in members
        and "__init__" not in method_list
    ):
        method_list = ["__init__"] + method_list

    for name in cls_dict.get("__annotations__", {}):
        if name not in attrs:
            attrs.append(name)
    return method_list, attrs


def _is_sqlalchemy_orm_model(cls):
    return nested(lambda: str(cls.__class__.__module__), "").startswith(
        "sqlalchemy.orm"
    )


def _get_sqlalchemy_init(attributes_with_type):
    res = [f"def __init__(self, *,"]
    for p, p_type in attributes_with_type:
        res.append(f"{INDENT}{p}: {p_type} = None,")
    res.append(f"{INDENT}**kw")
    res.append(f"): ...")
    return res


def _is_not_generic_and_private_class_or_module(the_type):
    if type_is_generic(the_type):
        return False
    return getattr(the_type, "__module__", "").startswith("_") or nested(
        lambda: the_type.__class__.__name__, ""
    ).startswith("_")


def _get_methods_info(
    cls, locals_attrs, additional_classes, ignore_attributes=None
) -> list:
    ignore_attributes = ignore_attributes or set()
    members = {}
    members.update(dict(cls.__dict__))
    annotations = cls.__dict__.get("__annotations__", {})
    for a in annotations:
        members[a] = annotations[a]
    members.update(annotations)

    method_list, cls_attrs_draft = _get_method_and_attr_list(cls, members)
    cls_attrs = [a for a in cls_attrs_draft if a not in ignore_attributes]
    attributes_with_type = []
    for attr in cls_attrs:
        the_type = members.get(attr, None)
        if _is_not_generic_and_private_class_or_module(the_type):
            continue
        if inspect.isclass(the_type) or type_is_generic(the_type):
            resolved_type = (
                get_type_info(the_type, locals_attrs, additional_classes)
                if the_type
                else "Any"
            )
        else:
            resolved_type = (
                get_type_info(the_type.__class__, locals_attrs, additional_classes)
                if the_type is not None
                else "Any"
            )
        attributes_with_type.append((attr, resolved_type))
    method_by_name = [
        f"{attr}: {resolved_type}" for (attr, resolved_type) in attributes_with_type
    ]
    cls_dict = cls.__dict__

    for name in method_list:
        method_cls = members[name].__class__
        is_property = False
        func = cls_dict.get(name) if name in cls_dict else getattr(cls, name, None)
        func = (
            getattr(cls, name)
            if isinstance(func, (classmethod, staticmethod, property))
            and not _is_sqlalchemy(func)
            else func
        )
        if isinstance(func, property):
            is_property = True
            func = func.__get__
        try:
            sig = inspect.signature(func)
            return_annotations = (
                ""
                if sig.return_annotation == inspect._empty
                else f" -> None"
                if sig.return_annotation is None
                else f" -> {get_type_info(sig.return_annotation, locals_attrs, additional_classes)}"
            )
            params_by_name = []
            if _is_sqlalchemy_orm_model(cls) and name == "__init__":
                method_by_name.extend(_get_sqlalchemy_init(attributes_with_type))
                continue

            if method_cls is classmethod:
                params_by_name.append(("cls", ""))
            if is_property:
                params_by_name.append(("self", ""))
            found_last_positional = False
            arg_position = 0
            for p, v in sig.parameters.items():
                if is_property and arg_position < 2:
                    continue
                arg_position += 1
                optional_globe = (
                    "**"
                    if v.kind == inspect.Parameter.VAR_KEYWORD
                    else "*"
                    if v.kind == inspect.Parameter.VAR_POSITIONAL
                    else ""
                )
                if v.kind == inspect.Parameter.VAR_POSITIONAL:
                    found_last_positional = True
                if (
                    v.kind == inspect.Parameter.KEYWORD_ONLY
                    and not found_last_positional
                ):
                    params_by_name.append(("*", ""))
                    found_last_positional = True
                default = (
                    ""
                    if v.default == inspect._empty
                    else f" = {v.default.__name__}"
                    if inspect.isclass(v.default)
                    else " = None"
                )
                type_annotation = (
                    ""
                    if v.annotation == inspect._empty
                    else f": {get_type_info(v.annotation, locals_attrs, additional_classes)}"
                )
                p_name = f"{optional_globe}{p}"
                type_annotation = (
                    type_annotation[: -len(" = None")]
                    if (type_annotation.endswith("= None") and default)
                    else type_annotation
                )
                params_by_name.append((p_name, f"{type_annotation}{default}"))

            params_as_str = ", ".join([f"{k}{v}" for (k, v) in params_by_name])
            method_by_name.append("")
            if method_cls is staticmethod:
                method_by_name.append("@staticmethod")
            elif method_cls is classmethod:
                method_by_name.append("@classmethod")
            elif is_property:
                method_by_name.append("@property")
            method_by_name.append(
                f"def {name}({params_as_str}){return_annotations}: ..."
            )
            if name == "__init__" and not issubclass(cls, Structure):
                try:
                    source = inspect.getsource(members[name])
                    init_type_by_attr = extract_attributes_from_init(
                        source, locals_attrs, additional_classes
                    )
                    for attr_name, attr_type in init_type_by_attr.items():
                        if attr_name not in dict(attributes_with_type):
                            method_by_name = [
                                f"{attr_name}: {attr_type}"
                            ] + method_by_name
                except:
                    logging.info("not __init__ implementation found")
        except Exception as e:
            logging.warning(e)
            method_by_name.append(f"def {name}(self, *args, **kw): ...")

    return method_by_name


def _get_init(cls, ordered_args: dict, additional_properties_default: bool) -> str:
    init_params = f",\n{INDENT * 2}".join(
        [f"{INDENT * 2}self"] + [f"{k}: {v}" for k, v in ordered_args.items()]
    )
    kw_opt = (
        f",\n{INDENT * 2}**kw"
        if getattr(cls, "_additionalProperties", additional_properties_default)
        else ""
    )
    return f"    def __init__(\n{init_params}{kw_opt}\n{INDENT}): ..."


def _get_additional_structure_methods(
    cls, ordered_args: dict, additional_properties_default: bool
) -> str:
    ordered_args_with_none = {}
    for k, v in ordered_args.items():
        ordered_args_with_none[k] = v if v.endswith("= None") else f"{v} = None"
    params = [f"{k}: {v}" for k, v in ordered_args_with_none.items()]
    params_with_self = f",\n{INDENT * 2}".join([f"{INDENT * 2}self"] + params)
    kw_opt = (
        f",\n{INDENT * 2}**kw"
        if getattr(cls, "_additionalProperties", additional_properties_default)
        else ""
    )
    shallow_clone = f"    def shallow_clone_with_overrides(\n{params_with_self}{kw_opt}\n{INDENT}): ..."

    params_with_cls = f",\n{INDENT*2}".join(
        [f"{INDENT}cls", "source_object: Any", "*"] + params
    )

    from_other_class = f"\n{INDENT}".join(
        [
            "",
            "@classmethod",
            "def from_other_class(",
            f"{params_with_cls}{kw_opt}\n{INDENT}): ...",
        ]
    )
    return "\n".join([shallow_clone, "", from_other_class])


def get_stubs_of_structures(
    struct_classe_by_name: dict,
    local_attrs,
    additional_classes,
    additional_properties_default,
) -> list:
    out_src = []
    for cls_name, cls in struct_classe_by_name.items():
        fields_info = get_all_type_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        method_info = _get_methods_info(
            cls,
            locals_attrs=local_attrs,
            additional_classes=additional_classes,
            ignore_attributes=set(fields_info.keys()),
        )

        ordered_args = _get_ordered_args(fields_info)
        bases = _get_bases_for_structure(cls, local_attrs, additional_classes)
        bases_str = f"({', '.join(bases)})" if bases else ""
        out_src.append(f"class {cls_name}{bases_str}:")
        if not fields_info and not method_info:
            out_src.append(f"{INDENT}pass")
            out_src.append("")
            continue

        if not [m for m in method_info if m.startswith("def __init__(self")]:
            out_src.append(_get_init(cls, ordered_args, additional_properties_default))
        out_src.append("")
        out_src.append(
            _get_additional_structure_methods(
                cls, ordered_args, additional_properties_default
            )
        )
        out_src.append("")
        out_src.append("")

        for field_name, type_name in ordered_args.items():
            out_src.append(f"    {field_name}: {type_name}")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def get_stubs_of_enums(
    enum_classes_by_name: dict, local_attrs, additional_classes
) -> list:
    out_src = []
    for cls_name, cls in enum_classes_by_name.items():

        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        bases = _get_bases(cls, local_attrs, additional_classes)
        bases_str = f"({', '.join(bases)})" if bases else ""
        out_src.append(f"class {cls_name}{bases_str}:")

        enum_values = [f"{INDENT}{v.name} = enum.auto()" for v in cls] or [
            f"{INDENT}pass"
        ]
        out_src.extend(enum_values)
        out_src.append("")

        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def add_imports(local_attrs: dict, additional_classes, existing_imports: set) -> list:
    base_typing = ["Union", "Optional", "Any", "TypeVar", "Type", "NoReturn"]
    typing_types_to_import = [t for t in base_typing if t not in existing_imports]
    base_import_statements = []
    if typing_types_to_import:
        base_import_statements.append(
            f"from typing import {', '.join(typing_types_to_import)}"
        )
    base_import_statements += [
        "from typedpy import Structure",
        "",
    ]
    extra_imports_by_name = _get_mapped_extra_imports(additional_classes)
    extra_imports = {
        f"from {_get_package(v, local_attrs)} import {k}{_as_something(k, local_attrs)}"
        for k, v in extra_imports_by_name.items()
        if (
            (
                k not in local_attrs
                or local_attrs[k].__module__ != local_attrs["__name__"]
            )
            and k not in existing_imports
        )
    }
    return base_import_statements + sorted(extra_imports)


def _get_enum_classes(attrs, only_calling_module):
    res = {}
    for k, v in attrs.items():
        if (
            inspect.isclass(v)
            and issubclass(v, enum.Enum)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        ):
            res[k] = v

    return res


def _get_other_classes(attrs, only_calling_module):
    res = {}
    for k, v in attrs.items():
        if (
            inspect.isclass(v)
            and not issubclass(v, enum.Enum)
            and not issubclass(v, Structure)
        ):
            res[k] = v

    return res


def _get_functions(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isfunction(v)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_type_annotation(
    prefix: str, annotation, default: str, local_attrs, additional_classes
):
    def _correct_for_return_annotation(res: str):
        if "->" in prefix:
            return res[:-7] if res.endswith("= None") else res
        if default:
            cleaned_res = res[: -len(" = None")] if res.endswith("= None") else res
            return f"{cleaned_res}{default}"
        return res

    try:
        res = (
            ""
            if annotation == inspect._empty
            else f"{prefix}{get_type_info(annotation, local_attrs, additional_classes)}"
        )
        return _correct_for_return_annotation(res)
    except Exception as e:
        logging.exception(e)
        return ""


def get_stubs_of_functions(func_by_name, local_attrs, additional_classes) -> list:
    def _convert_default(d):
        return d.__name__ if inspect.isclass(d) else None

    out_src = []
    for name, func in func_by_name.items():
        sig = inspect.signature(func)
        return_annotations = _get_type_annotation(
            " -> ", sig.return_annotation, "", local_attrs, additional_classes
        )
        params_by_name = []
        found_last_positional = False
        for p, v in sig.parameters.items():
            default = (
                ""
                if v.default == inspect._empty
                else f" = {_convert_default(v.default)}"
            )
            type_annotation = _get_type_annotation(
                ": ", v.annotation, default, local_attrs, additional_classes
            )
            optional_globe = (
                "**"
                if v.kind == inspect.Parameter.VAR_KEYWORD
                else "*"
                if v.kind == inspect.Parameter.VAR_POSITIONAL
                else ""
            )
            if v.kind == inspect.Parameter.VAR_POSITIONAL:
                found_last_positional = True
            if v.kind == inspect.Parameter.KEYWORD_ONLY and not found_last_positional:
                params_by_name.append(("*", ""))
                found_last_positional = True
            p_name = f"{optional_globe}{p}"
            params_by_name.append((p_name, type_annotation))
        params_as_str = ", ".join([f"{k}{v}" for (k, v) in params_by_name])

        out_src.append(f"def {name}({params_as_str}){return_annotations}: ...")
        out_src.append("\n")
    return out_src


def _get_bases(cls, local_attrs, additional_classes) -> list:
    res = []
    for b in cls.__bases__:
        if b is object or b.__module__ == "typing" and b is not typing.Generic:
            continue
        if not _is_sqlalchemy(b):
            the_type = get_type_info(b, local_attrs, additional_classes)
            if b is typing.Generic and the_type == "Generic":
                params = [p.__name__ for p in cls.__parameters__]
                params_st = f"[{', '.join(params)}]" if params else ""
                res.append(f"{the_type}{params_st}")

            elif the_type != "Any":
                res.append(the_type)
    return res


def _get_bases_for_structure(cls, local_attrs, additional_classes) -> list:
    res = []
    for b in cls.__bases__:
        if b is object or b.__module__ == "typing":
            continue
        if b.__module__.startswith("typedpy"):
            continue
        the_type = get_type_info(b, local_attrs, additional_classes)
        if the_type != "Any":
            res.append(the_type)
    res.append("Structure")
    return res


def get_stubs_of_other_classes(
    *, other_classes, local_attrs, additional_classes, additional_imports
):
    out_src = []
    for cls_name, cls in other_classes.items():
        if cls.__module__ != local_attrs["__name__"]:
            if cls_name not in additional_imports:
                out_src += [f"class {cls_name}:", f"{INDENT}pass", ""]
            continue

        bases = _get_bases(cls, local_attrs, additional_classes)
        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        bases_str = f"({', '.join(bases)})" if bases else ""
        out_src.append(f"class {cls_name}{bases_str}:")
        if not method_info:
            out_src.append(f"{INDENT}pass")
        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def get_typevars(attrs, additional_classes):
    res = []
    for k, v in attrs.items():
        if isinstance(v, typing.TypeVar):
            constraints = ", ".join(
                [get_type_info(x, attrs, additional_classes) for x in v.__constraints__]
            )
            constraints_s = f", {constraints}" if constraints else ""
            res.append(f'{k} = TypeVar("{k}"{constraints_s})')
    return [""] + res + [""]


def create_pyi(
    calling_source_file,
    attrs: dict,
    only_current_module: bool = True,
    additional_properties_default=True,
):
    full_path: Path = Path(calling_source_file)
    pyi_path = (full_path.parent / f"{full_path.stem}.pyi").resolve()
    out_src = []
    additional_imports = []
    imported = list(get_imports(attrs.get("__file__")))
    if only_current_module:
        for level, pkg_name, val_info, alias in imported:
            level_s = "." * level
            val = ".".join(val_info)
            alias = alias or val
            if val and pkg_name:
                if val != "*":
                    out_src.append(f"from {level_s}{pkg_name} import {val} as {alias}")
                    additional_imports.append(alias)
                else:
                    out_src.append(f"from {level_s}{pkg_name} import {val}")
            elif pkg_name:
                out_src.append(f"import {level_s}{pkg_name}")
                additional_imports.append(pkg_name)
            else:
                out_src.append(f"import {level_s}{alias}")
                additional_imports.append(alias)

    enum_classes = _get_enum_classes(attrs, only_calling_module=only_current_module)
    if enum_classes and enum not in additional_imports:
        out_src += ["import enum", ""]
    struct_classes = _get_struct_classes(attrs, only_calling_module=only_current_module)
    other_classes = _get_other_classes(attrs, only_calling_module=only_current_module)
    functions = _get_functions(attrs, only_calling_module=only_current_module)

    additional_classes = set()
    out_src += get_typevars(attrs, additional_classes=additional_classes)

    out_src += _get_consts(
        attrs,
        additional_classes=additional_classes,
        additional_imports=additional_imports,
    )

    out_src += get_stubs_of_enums(
        enum_classes, local_attrs=attrs, additional_classes=additional_classes
    )
    out_src += get_stubs_of_other_classes(
        other_classes=other_classes,
        local_attrs=attrs,
        additional_classes=additional_classes,
        additional_imports=additional_imports,
    )
    out_src += get_stubs_of_structures(
        struct_classes,
        local_attrs=attrs,
        additional_classes=additional_classes,
        additional_properties_default=additional_properties_default,
    )

    out_src += get_stubs_of_functions(
        functions, local_attrs=attrs, additional_classes=additional_classes
    )

    from_future_import = [
        (i, s) for i, s in enumerate(out_src) if s.startswith("from __future__")
    ]
    for number_of_deletions, (i, s) in enumerate(from_future_import):
        out_src = (
            out_src[: i - number_of_deletions] + out_src[i + 1 - number_of_deletions :]
        )
    out_src = (
        [x[1] for x in from_future_import]
        + add_imports(
            local_attrs=attrs,
            additional_classes=additional_classes,
            existing_imports=set(additional_imports),
        )
        + out_src
    )

    out_src = AUTOGEN_NOTE + out_src
    out_s = "\n".join(out_src)
    with open(pyi_path, "w", encoding="UTF-8") as f:
        f.write(out_s)


def _get_consts(attrs, additional_classes, additional_imports):
    def _is_of_builtin(v) -> bool:
        return isinstance(
            v, (int, float, str, dict, list, set, complex, bool, frozenset)
        )

    def _as_builtin(v) -> str:
        if isinstance(v, str):
            return ""
        return v if isinstance(v, (int, float, complex, bool)) else v.__class__()

    res = []
    annotations = attrs.get("__annotations__", None) or {}
    constants = {
        k: v
        for (k, v) in attrs.items()
        if (
            _is_of_builtin(v)
            or v is None
            or k in annotations
            or (
                not inspect.isclass(v)
                and not inspect.isfunction(v)
                and not isinstance(v, typing.TypeVar)
            )
        )
        and not k.startswith("__")
        and k not in additional_imports
    }
    for c in constants:
        the_type = (
            get_type_info(annotations[c], attrs, additional_classes)
            if c in annotations
            else get_type_info(attrs[c].__class__, attrs, additional_classes)
            if not inspect.isclass(attrs[c]) and attrs[c] is not None
            else None
        )
        type_str = f": {the_type}" if the_type else ""
        val = (
            str(doublewrap_val(_as_builtin(attrs[c])))
            if _is_of_builtin(attrs[c])
            else "None"
            if attrs[c] is None
            else ""
        )
        val_st = f" = {val}" if val else ""
        res.append(f"{c}{type_str}{val_st}")
        res.append("")
    return res


def create_stub_for_file(
    abs_module_path: str,
    src_root: str,
    stubs_root: str = None,
    *,
    additional_properties_default=True,
):
    ext = os.path.splitext(abs_module_path)[-1].lower()
    if ext != ".py":
        return
    stem = Path(abs_module_path).stem
    dir_name = str(Path(abs_module_path).parent)
    relative_dir = relpath(dir_name, src_root)
    package_name = ".".join(Path(relative_dir).parts)
    module_name = stem if stem != "__init__" else package_name
    sys.path.append(str(Path(dir_name).parent))
    sys.path.append(src_root)
    spec = importlib.util.spec_from_file_location(module_name, abs_module_path)
    the_module = importlib.util.module_from_spec(spec)
    if not the_module.__package__:
        the_module.__package__ = package_name
    spec.loader.exec_module(the_module)

    pyi_dir = (
        Path(stubs_root) / Path(relative_dir)
        if stubs_root
        else Path(abs_module_path).parent
    )
    pyi_dir.mkdir(parents=True, exist_ok=True)
    (pyi_dir / Path("__init__.pyi")).touch(exist_ok=True)

    pyi_path = (pyi_dir / f"{stem}.pyi").resolve()
    if not getattr(the_module, "__package__", None):
        the_module.__package__ = ".".join(Path(relative_dir).parts)
    create_pyi(
        str(pyi_path),
        the_module.__dict__,
        additional_properties_default=additional_properties_default,
    )


def create_pyi_ast(calling_source_file, pyi_path):
    out_src = [
        "import datetime",
        "from typing import Optional, Any, Iterable, Union",
        "from typedpy import Structure",
    ]
    additional_imports = []
    imported = list(get_imports(calling_source_file))
    found_sqlalchmy = False
    for level, pkg_name, val_info, alias in imported:
        level_s = "." * level
        val = ".".join(val_info)
        alias = alias or val
        if pkg_name and pkg_name.startswith("sqlalchemy"):
            found_sqlalchmy = True
        if val and pkg_name:
            if val != "*":
                out_src.append(f"from {level_s}{pkg_name} import {val} as {alias}")
                additional_imports.append(alias)
            else:
                out_src.append(f"from {level_s}{pkg_name} import {val}")
        elif pkg_name:
            out_src.append(f"import {level_s}{pkg_name}")
            additional_imports.append(pkg_name)
        else:
            out_src.append(f"import {level_s}{alias}")
            additional_imports.append(alias)

    if found_sqlalchmy:
        out_src.extend(["from sqlalchemy import Column"])
        out_src.extend([""] * 3)
        models, functions = get_models(calling_source_file)
        out_src += models_to_src(models)
        out_src += functions_to_str(functions)
    out_src = AUTOGEN_NOTE + out_src
    out_s = "\n".join(out_src)
    with open(pyi_path, "w", encoding="UTF-8") as f:
        f.write(out_s)


def create_stub_for_file_using_ast(
    abs_module_path: str, src_root: str, stubs_root: str = None, **kw
):
    ext = os.path.splitext(abs_module_path)[-1].lower()
    if ext != ".py":
        return
    stem = Path(abs_module_path).stem
    dir_name = str(Path(abs_module_path).parent)
    relative_dir = relpath(dir_name, src_root)

    pyi_dir = (
        Path(stubs_root) / Path(relative_dir)
        if stubs_root
        else Path(abs_module_path).parent
    )
    pyi_dir.mkdir(parents=True, exist_ok=True)
    (pyi_dir / Path("__init__.pyi")).touch(exist_ok=True)

    pyi_path = (pyi_dir / f"{stem}.pyi").resolve()
    create_pyi_ast(abs_module_path, str(pyi_path))



from typedpy.structures import TypedField, _FieldMeta


class _SubClassMeta(_FieldMeta):
    def __getitem__(cls, value):
        return cls(clazz=value)  # pylint: disable=E1120, E1123


class SubClass(TypedField, metaclass=_SubClassMeta):
    """
    A Subclass of an given class

    Arguments:
        clazz(type):
            The class that the field is subclass of

            .. code-block:: python

               class Foo(Structure): pass
               class Bar(Foo): pass

               class Container(Structure):
                   data: dict[SubClass(clazz=Foo), str]

               container = Container(data={Bar: "bar"})
    """

    _ty = type

    def __init__(
        self,
        *args,
        clazz: type,
        **kwargs,
    ):
        if not isinstance(clazz, type):
            raise TypeError("SubClass must accept a class type as argument")
        self._clazz = clazz
        super().__init__(*args, **kwargs)

    def _validate(self, value):
        if not issubclass(value, self._clazz):
            raise TypeError(
                f"{self._name}: Expected a subclass of {self._clazz.__name__}; Got {value}"
            )

    def __set__(self, instance, value):
        if not getattr(instance, "_skip_validation", False):
            self._validate(value)
        super().__set__(instance, value)



import ast
import typing

from typedpy import Array, Map, Structure
from typedpy.type_info_getter import get_type_info

INDENT = " " * 4


def get_imports(path):
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    for node in ast.iter_child_nodes(root):
        level = 0
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module
            level = node.level
        else:
            continue

        for n in node.names:
            yield (level, module, n.name.split("."), n.asname)


class FunctionInfo(Structure):
    name: str
    positional_args: Array[str]
    keyword: typing.Optional[str]
    keyword_only_args: Array[str]
    returns: str
    decorators: Array[str]

    _ignore_none = True


class ModelClass(Structure):
    name: str
    bases: Array[str]
    columns: Map[str, str]
    relationships: Map[str, str]
    functions: Array[FunctionInfo]

    @staticmethod
    def type_by_sqlalchemy_type(sqlalchemy_type: str) -> str:
        the_type = sqlalchemy_type.split(".")[-1]
        return {
            "BigInteger": "int",
            "Integer": "int",
            "Boolean": "bool",
            "String": "str",
            "DateTime": "datetime.datetime",
            "Date": "datetime.date",
            "Float": "float",
            "Interval": "datetime.timedelta",
            "Time": "datetime.time",
        }.get(the_type, "Any")


def _get_param_type(ast_type):
    if isinstance(ast_type, ast.Name):
        return ast_type.id
    elif isinstance(ast_type, ast.Subscript):
        value = ast_type.value.id
        if isinstance(ast_type.slice, ast.Tuple):
            args = [_get_param_type(x) for x in ast_type.slice.dims]
        else:
            args = [ast_type.slice.id]
        return f"{value}[{', '.join(args)}]"


def _extract_kw_args(args_info) -> str:
    if args_info.kwarg:
        return f"**{args_info.kwarg.arg}"
    return ""


def _extract_kwonly_arg(arg, default) -> str:
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    default_s = " = None" if default is not None else ""
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    return f"{arg.arg}{the_type}{default_s}"


def _extract_arg(arg):
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    return f"{arg.arg}{the_type}"


def method_to_src(func: FunctionInfo):
    res = []
    for d in func.decorators:
        res.append(f"@{d}")

    res.append(f"def {func.name}(")
    for a in func.positional_args:
        res.append(f"{INDENT}{a},")
    if func.keyword_only_args:
        res.append(f"{INDENT}*,")
        for a in func.keyword_only_args:
            res.append(f"{INDENT}{a},")
    if func.keyword:
        res.append(f"{INDENT}{func.keyword},")
    returns_s = f" -> {func.returns}" if func.returns else ""
    res.append(f"){returns_s}: ...")
    res.append("")
    return res


def functions_to_str(functions: typing.Iterable[FunctionInfo]) -> typing.Iterable[str]:
    res = []
    for func in functions:
        res.extend(method_to_src(func))
        res.append("")
        res.append("")

    return res


def models_to_src(models: typing.Iterable[ModelClass]) -> typing.Iterable[str]:
    res = []
    for model in models:
        bases = f"({', '.join(model.bases)})" if model.bases else ""
        res.append(f"class {model.name}{bases}:")
        column_with_type = [
            f"{INDENT}{field}: Union[Column, {the_type}]"
            for (field, the_type) in model.columns.items()
        ]
        fields_with_type = [
            f"{INDENT}{field}: {the_type}"
            for (field, the_type) in model.columns.items()
        ]
        for field in column_with_type:
            res.append(field)
        for rel in model.relationships:
            res.append(f"{INDENT}{rel}: Any")
        fields_arg_list = [f"{INDENT * 2}{f} = None," for f in fields_with_type]
        relationships_arg_list = [
            f"{INDENT * 3}{f} = None," for f in model.relationships
        ]
        res.append(f"{INDENT}def __init__(self,")
        res.extend(fields_arg_list)
        res.extend(relationships_arg_list)
        res.append(f"{INDENT}): ...")
        res.append("")
        if "Mappable" in bases:
            res.append(f"{INDENT}@classmethod")
            res.append(f"{INDENT}def from_structure(cls,")
            res.append(f"{INDENT*3}structure: Structure,")
            res.append(f"{INDENT*3}*,")
            res.append(f"{INDENT*3}ignore_props: list[str] = None,")
            res.extend(fields_arg_list)
            res.extend(relationships_arg_list)
            res.append(f"{INDENT}) -> {model.name}: ...")
            res.append("")
        for func in model.functions:
            res.extend([f"{INDENT}{line}" for line in method_to_src(func)])
            res.append("")
        res.append("")

    return res


def _get_function_info(node) -> FunctionInfo:
    decorators = [d.id for d in node.decorator_list]
    return_type = _get_param_type(node.returns) if node.returns else ""
    args = [_extract_arg(a) for a in node.args.args]
    keyword_args = _extract_kw_args(node.args)
    keyword_only_args = [
        _extract_kwonly_arg(a, d)
        for a, d in zip(node.args.kwonlyargs, node.args.kw_defaults)
    ]
    return FunctionInfo(
        name=node.name,
        positional_args=args,
        keyword=keyword_args,
        keyword_only_args=keyword_only_args,
        returns=return_type,
        decorators=decorators,
    )


def get_models(
    path,
) -> typing.Tuple[typing.Iterable[ModelClass], typing.Iterable[FunctionInfo]]:
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    classes = []
    functions = []
    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.FunctionDef):
            functions.append(_get_function_info(node))
        elif isinstance(node, ast.ClassDef):
            if node.bases:
                bases = [x.id for x in node.bases if isinstance(x, ast.Name)]
                if "Base" in bases:
                    bases.remove("Base")
                    class_name = node.name
                    body = node.body
                    methods = []
                    relationships = {}
                    columns = {}
                    for b in body:
                        if isinstance(b, ast.Assign) and isinstance(b.value, ast.Call):
                            field_name = b.targets[0].id
                            function_name = b.value.func.id
                            if function_name == "Column":
                                args = b.value.args
                                first = args[0]
                                column_type = (
                                    first.id
                                    if isinstance(first, ast.Name)
                                    else first.func.id
                                )
                                columns[
                                    field_name
                                ] = ModelClass.type_by_sqlalchemy_type(column_type)
                            if function_name == "relationship":
                                args = b.value.args
                                first = args[0]
                                entity_type = (
                                    first.id
                                    if isinstance(first, ast.Name)
                                    else first.value
                                )
                                relationships[field_name] = entity_type
                        if isinstance(b, ast.FunctionDef):
                            methods.append(_get_function_info(b))
                    classes.append(
                        ModelClass(
                            name=class_name,
                            bases=bases,
                            columns=columns,
                            relationships=relationships,
                            functions=methods,
                        )
                    )
    return classes, functions


def extract_attributes_from_init(source, locals_attrs, additional_classes):
    def _get_normalized_dict(args):
        return dict([x if len(x) == 2 else x + ["Any"] for x in args])

    root = ast.parse(source.strip())
    info = _get_function_info(root.body[0])
    self_name = info.positional_args[0]
    positional_args = [x.split(":") for x in info.positional_args[1:]]
    normalized_positional_args = _get_normalized_dict(positional_args)
    keyword_only_args_without_default = [
        x.split("=")[0] for x in info.keyword_only_args
    ]
    keyword_only_args = [x.split(":") for x in keyword_only_args_without_default]
    normalized_keyword_only_args = _get_normalized_dict(keyword_only_args)
    type_by_param = {**normalized_positional_args, **normalized_keyword_only_args}

    func_body = root.body[0].body

    res = {}
    for node in func_body:
        if isinstance(node, ast.Assign):
            if (
                isinstance(node.targets[0], ast.Attribute)
                and isinstance(node.targets[0].value, ast.Name)
                and node.targets[0].value.id == self_name
            ):
                # assignment to attribute
                attribute_name = node.targets[0].attr
                if isinstance(node.value, ast.Name) and node.value.id in type_by_param:
                    res[attribute_name] = type_by_param[node.value.id]
                else:
                    try:
                        value = eval(ast.unparse(node.value), locals_attrs)
                        the_type = get_type_info(
                            value.__class__, locals_attrs, additional_classes
                        )
                        res[attribute_name] = the_type
                    except:
                        res[attribute_name] = "Any"

        elif isinstance(node, ast.AnnAssign):
            annotation = ast.unparse(node.annotation)
            attribute_name = node.target.attr
            res[attribute_name] = annotation

    return res



"""
Module for custom deserialization mappers aggregation
"""
from collections.abc import Mapping
from enum import Enum, auto

from .commons import deep_get
from .structures import ClassReference, Field, Structure, SERIALIZATION_MAPPER
from .fields import Array, FunctionCall, Set, StructureReference


class Constant:
    """
    Mark a value as constant in a versioned mapper.
    This is useful if an attribute did not exist in a previous version
    and you want to assign it to a default value in the more recent version.
    """

    def __init__(self, val):
        self._val = val

    def __call__(self, *args, **kwargs):
        return self._val


class Deleted:
    """
    Used to mark an attribute as "removed" in a versioned mapper.
    This is unsupported (nor needed) in a "regular" serialization mapper.
    """

    pass


# pylint:disable=protected-access, missing-function-docstring, invalid-name
def _set_base_mapper_no_op(cls, for_serialization):
    mapper = {}
    for k, field_def in cls.get_all_fields_by_name().items():
        if isinstance(field_def, ClassReference):
            val = (
                aggregate_serialization_mappers(field_def._ty)
                if for_serialization
                else aggregate_deserialization_mappers(field_def._ty)
            )
            mapper[f"{k}._mapper"] = val
        elif isinstance(field_def, (Array, Set)):
            items = (
                [field_def.items]
                if isinstance(field_def.items, Field)
                else field_def.items
                if isinstance(field_def.items, list)
                else []
            )
            values = {}
            for i in items:
                if isinstance(i, ClassReference):
                    val = (
                        aggregate_serialization_mappers(i._ty)
                        if for_serialization
                        else aggregate_deserialization_mappers(i._ty)
                    )
                    values.update(val)
                elif isinstance(i, StructureReference):
                    val = (
                        aggregate_serialization_mappers(i._newclass)
                        if for_serialization
                        else aggregate_deserialization_mappers(i._newclass)
                    )
                    values.update(val)
            if values:
                mapper[f"{k}._mapper"] = values

        elif isinstance(field_def, StructureReference):
            val = (
                aggregate_serialization_mappers(field_def._newclass)
                if for_serialization
                else aggregate_deserialization_mappers(field_def._newclass)
            )
            mapper[f"{k}._mapper"] = val
        mapper[k] = k

    return mapper


def _convert_to_camelcase(key):
    words = key.split("_")
    return words[0] + "".join(w.title() for w in words[1:])


def _convert_to_snakecase(key):
    return "".join(
        ["_" + char.lower() if char.isupper() else char for char in key]
    ).lstrip("_")


def _apply_mapper(
    latest_mapper, key, previous_mapper, for_serialization, is_self=False
):
    val = key if is_self else previous_mapper.get(key, key)
    if latest_mapper == mappers.TO_CAMELCASE:
        return _convert_to_camelcase(val)
    if latest_mapper == mappers.TO_LOWERCASE:
        return val.upper()
    latest_mapper_val = latest_mapper.get(val, val)
    if isinstance(latest_mapper_val, (FunctionCall,)):
        args = (
            [(deep_get(previous_mapper, k) or k) for k in latest_mapper_val.args]
            if latest_mapper_val.args
            else [previous_mapper.get(key)]
        )
        if for_serialization and previous_mapper.get(key) != key:
            raise NotImplementedError(
                "Combining functions and other mapping in a serialization mapper is unsupported"
            )
        return FunctionCall(func=latest_mapper_val.func, args=args)
    return latest_mapper.get(val, val)


def add_mapper_to_aggregation(latest_mapper, previous_mapper, for_serialization=False):
    result_mapper = {}  # copy.deepcopy(previous_mapper)
    if not isinstance(latest_mapper, (Mapping, mappers)):
        raise TypeError("Mapper must be a mapping")
    for k, v in previous_mapper.items():
        if isinstance(latest_mapper, dict) and v == latest_mapper.get(k):
            result_mapper[k] = previous_mapper.get(k)
        elif v is DoNotSerialize:
            result_mapper[k] = DoNotSerialize
        elif isinstance(v, str):
            result_mapper[k] = _apply_mapper(
                latest_mapper, k, previous_mapper, for_serialization=for_serialization
            )
        elif isinstance(v, dict):
            if not k.endswith("._mapper"):
                raise ValueError(
                    "Invalid mapper. To map nested values, use the format <key>._mapper: {...}"
                )
            field_name = k[0:-8]
            mapped_key = (
                field_name
                if for_serialization
                else _apply_mapper(
                    latest_mapper,
                    field_name,
                    previous_mapper,
                    for_serialization,
                    is_self=True,
                )
            )

            sub_mapper = (
                latest_mapper.get(
                    f"{mapped_key}._mapper", latest_mapper.get(f"{field_name}._mapper")
                )
                if not isinstance(latest_mapper, mappers)
                else latest_mapper
            )
            if sub_mapper:
                result_mapper[f"{mapped_key}._mapper"] = add_mapper_to_aggregation(
                    sub_mapper, v, for_serialization=for_serialization
                )
            else:
                result_mapper[f"{mapped_key}._mapper"] = v
        elif isinstance(v, FunctionCall):
            args = v.args or [k]
            mapped_args = (
                args
                if for_serialization
                else [
                    _apply_mapper(
                        latest_mapper,
                        a,
                        previous_mapper,
                        for_serialization=for_serialization,
                        is_self=(a == k),
                    )
                    for a in args
                ]
            )
            if isinstance(latest_mapper, Mapping) and isinstance(
                latest_mapper.get(k), FunctionCall
            ):
                raise NotImplementedError(
                    "Combining multiple functions in serialization is unsupported"
                )
            result_mapper[k] = FunctionCall(func=v.func, args=mapped_args)

    return result_mapper


class DoNotSerialize:
    pass


class mappers(Enum):
    """Useful custom mappers"""

    TO_LOWERCASE = auto()
    TO_CAMELCASE = auto()
    NO_MAPPER = auto()


def get_mapper(val: Structure):
    return getattr(val.__class__, SERIALIZATION_MAPPER, {})


def aggregate_deserialization_mappers(
    cls, override_mapper=None, camel_case_convert=False
):
    base_mapper = _set_base_mapper_no_op(cls, for_serialization=False)
    aggregate_mapper = base_mapper
    override_mapper = (
        override_mapper
        if isinstance(override_mapper, list)
        else [override_mapper]
        if override_mapper
        else None
    )
    mappers_list = (
        override_mapper
        if override_mapper
        else cls.get_aggregated_deserialization_mapper()
    )
    for m in mappers_list:
        aggregate_mapper = add_mapper_to_aggregation(m, aggregate_mapper)
    if camel_case_convert:
        aggregate_mapper = add_mapper_to_aggregation(
            mappers.TO_CAMELCASE, aggregate_mapper
        )
    return aggregate_mapper


def aggregate_serialization_mappers(
    cls, override_mapper=None, camel_case_convert=False
):
    base_mapper = _set_base_mapper_no_op(cls, for_serialization=True)
    aggregate_mapper = base_mapper
    override_mapper = (
        override_mapper
        if isinstance(override_mapper, list)
        else [override_mapper]
        if override_mapper
        else None
    )
    mappers_list = (
        override_mapper
        if override_mapper
        else cls.get_aggregated_serialization_mapper()
    )
    for m in mappers_list:
        aggregate_mapper = add_mapper_to_aggregation(m, aggregate_mapper, True)
    if camel_case_convert:
        aggregate_mapper = add_mapper_to_aggregation(
            mappers.TO_CAMELCASE, aggregate_mapper, True
        )
    return aggregate_mapper



import json
import sys
from collections.abc import Iterable, Mapping, Generator
from functools import reduce, wraps
from inspect import signature
from typing import Optional, Union

py_version = sys.version_info[0:2]
python_ver_36 = py_version == (3, 6)
python_ver_atleast_than_37 = py_version > (3, 6)
python_ver_atleast_39 = py_version >= (3, 9)
python_ver_atleast_310 = py_version >= (3, 10)


def wrap_val(v):
    return f"'{v}'" if isinstance(v, str) else v


def doublewrap_val(v):
    return f'"{v}"' if isinstance(v, str) else v


def _is_dunder(name):
    """Returns True if a __dunder__ name, False otherwise."""
    return (
        len(name) > 4
        and name[:2] == name[-2:] == "__"
        and name[2] != "_"
        and name[-3] != "_"
    )


def _is_sunder(name):
    """Returns True if a _sunder name, False otherwise."""
    return len(name) > 2 and name[0] == "_" and name[1:2] != "_"


class InvalidStructureErr(ValueError, TypeError):
    pass


def raise_errs_if_needed(cls, errors):
    if errors:
        cls_name = cls.__name__
        messages = [f"{cls_name}.{e}" for e in errors]
        raise InvalidStructureErr(json.dumps(messages)) from errors[0]


def first_in(my_list: Iterable, ignore_none: bool = False) -> Optional:
    """
    get the first in an Iterable (i.e. list, tuple, generator, dict, etc.).
    Optionally ignoring None values.

       Arguments:
           my_list(Iterable):
               The input iterable, such as a list
           ignore_none(bool): optional
               whether or not to ignore None values. Default is False


       Returns:
           The first value found in the input, or None if none found
    """
    if ignore_none:
        return next(filter(None, iter(my_list)), None)
    return next(iter(my_list), None)


def nested(func, default=None):
    """
    get a nested value if it exists, or return the given default if it doesn't

        Arguments:
            func(function):
                A function with no arguments that returns the nested value
            default:
                the default value, in case the nested value does not exist


        Returns:
            the nested attribute(s) or the default value. For example:

        For example:

            .. code-block:: python

               d = D(c=C(b=B(a=A(i=5))))
               assert nested(lambda: d.c.b.a.i) == 5

               d.c.foo = [1, 2, 3]
               assert nested(lambda: d.c.foo[100]) is None
               assert nested(lambda: d.c.foo[2]) == 3
    """
    try:
        return func()
    except (AttributeError, TypeError, IndexError, KeyError):
        return default


def flatten(iterable, ignore_none=False) -> list:
    """
    Flatten an iterable completely, getting rid of None values

       Arguments:
              iterable(Iterable):
                   the input
              ignore_none(bool): optional
                   whether to skip None values or not. Default is False.

       Returns:
           A flattened list

           .. code-block:: python

              flatten(
              [
                [[1]],
                [[2], 3, None, (5,)],
                []
              ], ignore_none=True) == [1, 2, 3, 5]

    """
    res = (
        sum([flatten(x, ignore_none) for x in iterable], [])
        if isinstance(iterable, (list, tuple, Generator))
        else [iterable]
    )
    if isinstance(res, (list, tuple, Generator)) and ignore_none:
        return [x for x in res if x is not None]
    return res


def deep_get(dictionary, deep_key, default=None, do_flatten=False):
    """
    Get a nested value from within a dictionary. Supports also nested lists, in
    which case the result is a a list of values

    Arguments:
           dictionary(dict):
                the input
           deep_key(str):
                nested key of the form aaa.bbb.ccc.ddd
           default:
                the default value, in case the path does not exist
           do_flatten(bool): optional
                flatten the outputs, in case the result is multiple outputs

    Returns:
        the nested attribute(s) or the default value. For example:

    For example:

        .. code-block:: python

           example = {"a": {"b": [{"c": [None, {"d": [1]}]}, {"c": [None, {"d": [2]}, {"d": 3}]}, {"c": []}]}}
           assert deep_get(example, "a.b.c.d") == [[[1]], [[2], 3], []]
           assert deep_get(example, "a.b.c.d", do_flatten=True) == [1, 2, 3]

    """

    def _get_next_level(d: Optional[Union[Mapping, Iterable]], key, default):
        if isinstance(d, Mapping):
            return d.get(key, default)
        if isinstance(d, (list, tuple, Generator)):
            res = [_get_next_level(r, key, default) for r in d if r is not None]
            return res
        return default

    keys = deep_key.split(".")
    result = reduce(
        lambda d, key: _get_next_level(d, key, default) if d else default,
        keys,
        dictionary,
    )
    if isinstance(result, (list, tuple, Generator)) and do_flatten:
        result = flatten(result)
    return result


def default_factories(func):
    """
     A function decorator that allows to have default values that are generators of the actual
     default values to be used. This is useful when the default values are mutable, like
     dicts or lists

        Arguments:
               iterable(Iterable):
                    the input
               ignore_none(bool): optional
                    whether to skip None values or not. Default is False.

        Returns:
            A flattened list

    For example:

            .. code-block:: python

               @default_factories
               def func(a, b = 0, c = list, d = dict):
                    return a, b, c, d

               assert func(1) == 1, 0, [] , {}

    """
    func_signature = signature(func, follow_wrapped=True)

    @wraps(func)
    def decorated(*args, **kwargs):
        bound = func_signature.bind(*args, **kwargs)

        for k, v in func_signature.parameters.items():
            if k not in bound.arguments:
                default = (
                    v.default()
                    if callable(v.default) and v.default != v.empty
                    else v.default
                )
                if v.default != v.empty:
                    kwargs[k] = default
        return func(*args, **kwargs)

    return decorated



import enum

from typedpy.structures import StructMeta


def keys_of(*enum_classes):
    """
    A class decorator that ensures the class fields include all the names of the enum values of the given enum classes.
     If not, the class definition throws a TypeError.
    For example:

    .. code-block:: python

        class Role(enum.Enum):
            admin = auto()
            manager = auto()
            sales = auto()
            engineer = auto()
            driver = auto()


       @keys_of(Role)
       class SalaryRules(Structure):
            admin: Range
            manager: Range
            sales: Range

            policies: list[Policy]

       # This class definiton will throw the following exception:
       # TypeError: SalaryRules: missing fields: driver, engineer

    """
    for clazz in enum_classes:
        if not issubclass(clazz, enum.Enum):
            raise TypeError(f"keys_of requires enum classes as parameters; Got {clazz}")

    def wrapper(clazz: StructMeta):
        field_names = set(clazz.get_all_fields_by_name().keys())
        expected_field_names = set()
        for c in enum_classes:
            expected_field_names |= set(e.name for e in c)
        if not expected_field_names <= field_names:
            missing_fields = sorted(expected_field_names - field_names)
            raise TypeError(
                f"{clazz.__name__}: missing fields: {', '.join(missing_fields)}"
            )
        return clazz

    return wrapper



import re
import json
from json import JSONDecodeError

from .structures import ImmutableStructure, Structure
from .fields import String, AnyOf, Array


class ErrorInfo(ImmutableStructure):
    field = String
    value = String
    problem = AnyOf[String, Array]
    _required = ["problem"]


display_type_by_type = {
    "int": "an integer number",
    "str": "a text value",
    "float": "a decimal number",
    "list": "an array",
}

_expected_class_pattern = re.compile(r"^Expected\s<class '(.*)'>$")


def _transform_class_to_readable(problem: str):
    match = _expected_class_pattern.match(problem)
    if match:
        return f"Expected {display_type_by_type.get(match.group(1), match)}"
    return problem


_pattern_for_typepy_validation_1 = re.compile(r"^([a-zA-Z0-9_.]+): Got ([^;]*); (.*)$")
_pattern_for_typepy_validation_2 = re.compile(r"^([a-zA-Z0-9_.]+):\s(.*); Got (.*)$")
_pattern_for_typepy_validation_3 = re.compile(r"^([a-zA-Z0-9_.]+):\s(.*)$")


def standard_readable_error_for_typedpy_exception(e: Exception, top_level=True):
    err_message = str(e)
    if Structure.failing_fast():
        return _standard_readable_error_for_typedpy_exception_internal(err_message)
    else:
        try:
            errs = json.loads(err_message)
            return [
                _standard_readable_error_for_typedpy_exception_internal(e) for e in errs
            ]
        except JSONDecodeError as ex:
            if not top_level:
                raise ex
            return [
                _standard_readable_error_for_typedpy_exception_internal(err_message)
            ]


def _standard_readable_error_for_typedpy_exception_internal(err_message: str):
    def try_expand(problem_str):
        if not Structure.failing_fast():
            try:
                return standard_readable_error_for_typedpy_exception(
                    Exception(problem_str), top_level=False
                )
            except Exception:
                pass
        return problem_str

    match = _pattern_for_typepy_validation_1.match(err_message)
    if match:
        problem = _transform_class_to_readable(match.group(3))
        return ErrorInfo(
            value=match.group(2), problem=try_expand(problem), field=match.group(1)
        )

    match = _pattern_for_typepy_validation_2.match(err_message)
    if match:
        problem = _transform_class_to_readable(match.group(2))
        return ErrorInfo(
            value=match.group(3), problem=try_expand(problem), field=match.group(1)
        )

    match = _pattern_for_typepy_validation_3.match(err_message)
    if match:
        field = match.group(1)
        problem = _transform_class_to_readable(match.group(2))
        return ErrorInfo(problem=try_expand(problem), field=field)
    return ErrorInfo(problem=err_message)


def get_simplified_error(err: str, as_list=True):
    try:
        res = json.loads(err)
        if isinstance(res, list):
            fixed = []
            for e in res:
                try:
                    key, *rest = e.split(":")
                    val = ":".join(rest)
                    corrected = get_simplified_error(val.strip(), as_list=False)
                    fixed.append(f"{key}: {corrected}")
                except Exception:  # noqa
                    fixed.append(e)

            return fixed if len(res) > 1 or as_list is True else fixed[0]
    except JSONDecodeError:
        pass
    return err



"""
Additional types of fields: datefield, datetime, timestring, DateString,
Hostname, etc.
"""
import json
from datetime import datetime, date
import re

from .commons import wrap_val
from .structures import TypedField
from .fields import SerializableField, String

EmailAddress = String(pattern=r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9]+$)")


class JSONString(String):
    """
    A string of a valid JSON
    """

    def __set__(self, instance, value):
        json.loads(value)
        super().__set__(instance, value)


class IPV4(String):
    """
    A string field of a valid IP version 4
    """

    _ipv4_re = re.compile(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")

    def __set__(self, instance, value):
        if IPV4._ipv4_re.match(value) and all(
            0 <= int(component) <= 255 for component in value.split(".")
        ):
            super().__set__(instance, value)
        else:
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; wrong format for IP version 4"
            )

    def to_json_schema(self) -> dict:
        return {"type": "string", "format": "ipv4"}

    @classmethod
    def from_json_schema(cls, schema: dict):
        return "IPV4()" if schema == {"type": "string", "format": "ipv4"} else None


class HostName(String):
    """
    A string field of a valid host name
    """

    _host_name_re = re.compile(r"^[A-Za-z0-9][A-Za-z0-9\.\-]{1,255}$")

    def __set__(self, instance, value):
        if not HostName._host_name_re.match(value):
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; wrong format for hostname"
            )
        components = value.split(".")
        for component in components:
            if len(component) > 63:
                raise ValueError(
                    f"{self._name}: Got {wrap_val(value)}; wrong format for hostname"
                )
        super().__set__(instance, value)

    def to_json_schema(self) -> dict:
        return {"type": "string", "format": "hostname"}

    @classmethod
    def from_json_schema(cls, schema: dict):
        return (
            "HostName()" if schema == {"type": "string", "format": "hostname"} else None
        )


class DateString(String):
    """
    A string field of the format '%Y-%m-%d' that can be converted to a date

    Arguments:
          date_format(str): optional
              an alternative date format

    """

    _ty = str

    def __init__(self, *args, date_format="%Y-%m-%d", **kwargs):
        self._format = date_format
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value):
        super().__set__(instance, value)
        try:
            datetime.strptime(value, self._format)
        except ValueError as ex:
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; {ex.args[0]}"
            ) from ex

    def to_json_schema(self) -> dict:
        return {"type": "string", "format": "date"}

    @classmethod
    def from_json_schema(cls, schema: dict):
        return (
            "DateString()" if schema == {"type": "string", "format": "date"} else None
        )


class TimeString(TypedField):
    """
    A string field of the format '%H:%M:%S' that can be converted to a time
    """

    _ty = str

    def __set__(self, instance, value):
        super().__set__(instance, value)
        try:
            datetime.strptime(value, "%H:%M:%S")
        except ValueError as ex:
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; {ex.args[0]}"
            ) from ex

    def to_json_schema(self) -> dict:
        return {"type": "string", "format": "time"}

    @classmethod
    def from_json_schema(cls, schema: dict):
        return (
            "TimeString()" if schema == {"type": "string", "format": "time"} else None
        )


class DateField(SerializableField):
    """
    A datetime.date field. Can accept either a date object, or a string
    that can be converted to a date, using the date_format in the constructor.

    Arguments:
         date_format(str): optional
             The date format used to convert to/from a string. Default is '%Y-%m-%d'

    Example:

    .. code-block:: python

        class Foo(Structure):
            date = DateField

        foo(date = date.today())
        foo(date = "2020-01-31")

    This is a SerializableField, thus can be serialized/deserialized.

    """

    def __init__(self, *args, date_format="%Y-%m-%d", **kwargs):
        self._date_format = date_format
        super().__init__(*args, **kwargs)

    def serialize(self, value):
        return value.strftime(self._date_format)

    def deserialize(self, value):
        try:
            return datetime.strptime(value, self._date_format).date()
        except ValueError as ex:
            raise ValueError(f"{self._name}: Got { wrap_val(value)}; {str(ex)}") from ex

    def __set__(self, instance, value):
        if isinstance(value, str):
            as_date = self.deserialize(value)
            super().__set__(instance, as_date)
        elif isinstance(value, datetime):
            super().__set__(instance, value.date())
        elif isinstance(value, date):
            super().__set__(instance, value)
        else:
            raise TypeError(
                f"{self._name}: Got {wrap_val(value)}; Expected date, datetime, or str"
            )

    @property
    def get_type(self):
        return date

    def to_json_schema(self) -> dict:
        return {"type": "string", "format": "date"}

    @classmethod
    def from_json_schema(cls, schema: dict):
        return "DateField()" if schema == {"type": "string", "format": "date"} else None


class DateTime(SerializableField):
    """
    A datetime.datetime field. Can accept either a datetime object, or a string
    that can be converted to a date, using the date_format in the constructor.

    Arguments:
        datetime_format(str): optional
            The format used to convert to/from a string. Default is '%m/%d/%y %H:%M:%S'

    Example:

       .. code-block:: python

           class Foo(Structure):
               timestamp = DateTime

           foo(timestamp = datetime.now())
           foo(timestamp = "01/31/20 07:15:45")

    This is a SerializableField, thus can be serialized/deserialized.

    """

    def __init__(self, *args, datetime_format="%m/%d/%y %H:%M:%S", **kwargs):
        self._datetime_format = datetime_format
        super().__init__(*args, **kwargs)

    def serialize(self, value: datetime):
        return value.strftime(self._datetime_format)

    def deserialize(self, value):
        try:
            if isinstance(value, int) and 2000000000 > value > 1000000000:
                return datetime.fromtimestamp(value)
            return datetime.strptime(value, self._datetime_format)
        except ValueError as ex:
            raise ValueError(f"{self._name}: Got {wrap_val(value)}; {str(ex)}") from ex

    def __set__(self, instance, value):
        if isinstance(value, datetime):
            super().__set__(instance, value)
        elif isinstance(value, (str, int)):
            as_datetime = self.deserialize(value)
            super().__set__(instance, as_datetime)
        else:
            raise TypeError(
                f"{self._name}: Got {wrap_val(value)}; Expected datetime or str"
            )

    @property
    def get_type(self):
        return datetime



import builtins
import collections.abc
import logging
import typing

from .enum import Enum
from .fields import AllOf, AnyOf, Map, OneOf
from .structures import Field, NoneField, get_typing_lib_info
from .utility import type_is_generic

builtins_types = [
    getattr(builtins, k)
    for k in dir(builtins)
    if isinstance(getattr(builtins, k), type)
]


def _get_anyof_typing(field, locals_attrs, additional_classes):
    union_fields = getattr(field, "_fields", [])
    if len(union_fields) == 2 and isinstance(union_fields[1], NoneField):
        info = get_type_info(union_fields[0], locals_attrs, additional_classes)
        return f"Optional[{info}] = None"

    fields = ",".join(
        [get_type_info(f, locals_attrs, additional_classes) for f in union_fields]
    )
    return f"Union[{fields}]"


def is_typeddict(the_type):
    return the_type.__class__.__name__ == "_TypedDictMeta"


def _get_type_info_for_typing_generic(
    the_type, locals_attrs, additional_classes
) -> typing.Optional[str]:
    origin = getattr(the_type, "__origin__", None)
    args = getattr(the_type, "__args__", [])
    if origin in {list, set, tuple}:
        if len(args) != 1:
            return origin.__name__
        return f"{origin.__name__}[{get_type_info(args[0], locals_attrs, additional_classes)}]"
    mapped_args = (
        [get_type_info(a, locals_attrs, additional_classes) for a in args]
        if args
        else []
    )

    if origin is collections.abc.Callable:
        args_st = (
            ""
            if not mapped_args
            else f"[{mapped_args[0]}]"
            if len(mapped_args) == 1
            else f"[[{','.join(mapped_args[:-1])}], {mapped_args[-1]}]"
        )
        additional_classes.add(typing.Callable)
        return f"Callable{args_st}"
    if origin is dict:
        if mapped_args:
            return f"dict[{', '.join(mapped_args)}]"
        return "dict"
    if origin is collections.abc.Iterable:
        additional_classes.add(typing.Iterable)
        args_st = "" if not mapped_args else f"[{mapped_args[0]}]"
        return f"Iterable{args_st}"

    if origin is collections.abc.Mapping:
        additional_classes.add(typing.Mapping)
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"Mapping{args_st}"

    if origin is type:
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"Type{args_st}"

    if origin is collections.abc.Iterator:
        additional_classes.add(typing.Iterator)
        args_st = "" if not mapped_args else f"[{mapped_args[0]}]"
        return f"Iterator{args_st}"

    if origin is typing.Union:
        if "None" in mapped_args:
            mapped_args = [a for a in mapped_args if str(a) != "None"]
            the_type = "Optional"
        else:
            the_type = "Type"
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"{the_type}{args_st}"

    if origin and mapped_args:
        mapped_origin = get_type_info(origin, locals_attrs, additional_classes)
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"{mapped_origin}{args_st}"
    return None


def _found_in_local_attrs(field, attrs):
    for k, v in attrs.items():
        if v is field:
            return k
    return None


def get_type_info(field, locals_attrs, additional_classes):
    try:
        if field is ...:
            return "..."
        found_import_key = _found_in_local_attrs(field, locals_attrs)
        if found_import_key:
            return found_import_key
        if isinstance(field, (list, dict)):
            cls_name = field.__class__.__name__
            mapped_args = [
                get_type_info(a, locals_attrs, additional_classes) for a in field
            ]
            args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
            return f"{cls_name}{args_st}"

        if field in [None, type(None)]:
            return "None"
        if isinstance(field, (AnyOf, OneOf, AllOf)):
            return _get_anyof_typing(field, locals_attrs, additional_classes)

        if isinstance(field, Map):
            if not field.items:
                return "dict"
            sub_types = ", ".join(
                [
                    get_type_info(f, locals_attrs, additional_classes)
                    for f in field.items
                ]
            )
            return f"dict[{sub_types}]"

        if isinstance(field, Enum):
            if field._is_enum:
                additional_classes.add(field._enum_class)
                return field._enum_class.__name__

        the_type = (
            getattr(field, "get_type", field) if isinstance(field, Field) else field
        )
        # deal with from __future__ import annotations
        if isinstance(the_type, str):
            the_type = eval(the_type, locals_attrs)
            field = the_type
            if isinstance(the_type, str):
                return the_type

        if is_typeddict(the_type):
            the_type = dict

        if the_type in builtins_types:
            return the_type.__name__

        if the_type in [typing.Any, typing.Optional, typing.Union, typing.NoReturn]:
            return getattr(the_type, "_name")
        if (
            getattr(the_type, "__module__", None)
            and the_type.__module__ != "builtins"
            and the_type not in locals_attrs
            and not type_is_generic(the_type)
        ):
            additional_classes.add(field)

        if type_is_generic(the_type):
            res = _get_type_info_for_typing_generic(
                the_type, locals_attrs, additional_classes
            )
            if res:
                return res

            return get_type_info(
                get_typing_lib_info(field), locals_attrs, additional_classes
            )

        return the_type.__name__
    except Exception as e:
        logging.exception(e)
        return "Any"


def get_all_type_info(cls, locals_attrs, additional_classes) -> dict:
    type_by_name = {}
    required = getattr(cls, "_required", None)
    for field_name, field in cls.get_all_fields_by_name().items():
        try:
            type_info_str: str = get_type_info(field, locals_attrs, additional_classes)
            if (
                field_name not in required
                and required is not None
                and not type_info_str.startswith("Optional[")
            ):
                type_info_str = f"Optional[{type_info_str}] = None"
            type_by_name[field_name] = type_info_str
        except Exception as e:
            logging.exception(e)

    return type_by_name



import enum
import logging
from collections import OrderedDict
from typing import Union

from .commons import default_factories, first_in, wrap_val
from .fields import (
    FunctionCall,
    Map,
    Negative,
    NonNegative,
    NonPositive,
    Positive,
    StructureReference,
    Integer,
    Number,
    Float,
    Array,
    String,
    ClassReference,
    Field,
    Boolean,
    AllOf,
    OneOf,
    AnyOf,
    NotField,
    Tuple,
    Set,
)
from .enum import Enum

from .extfields import DateString
from .mappers import Constant, DoNotSerialize, aggregate_serialization_mappers
from .structures import ADDITIONAL_PROPERTIES, NoneField, Structure

SCHEMA_PATTERN_PROPERTIES = "patternProperties"
SCHEMA_ADDITIONAL_PROPERTIES = "additionalProperties"
SCHEMA_PROPETIES = "properties"
SCHEMA_PROPERTY_NAMES = "propertyNames"


def get_mapper(field_cls):
    field_type_to_mapper = {
        StructureReference: StructureReferenceMapper,
        Integer: IntegerMapper,
        Number: NumberMapper,
        Float: FloatMapper,
        Array: ArrayMapper,
        Boolean: BooleanMapper,
        Enum: EnumMapper,
        String: StringMapper,
        DateString: DateStringMapper,
        AllOf: AllOfMapper,
        AnyOf: AnyOfMapper,
        OneOf: OneOfMapper,
        NotField: NotFieldMapper,
        Tuple: ArrayMapper,
        Set: ArrayMapper,
        Map: MapMapper,
    }
    for cls in field_cls.__mro__:
        if issubclass(cls, Field) and cls in field_type_to_mapper:
            return field_type_to_mapper[cls]
    raise NotImplementedError(f"schema mapping is not implemented for {field_cls}")


def _map_class_reference(reference, definitions_schema):
    definition, _ = structure_to_schema(getattr(reference, "_ty"), definitions_schema)
    name = getattr(reference, "_ty").__name__
    definitions_schema[name] = definition
    return {"$ref": f"#/definitions/{name}"}


def convert_to_schema(field, definitions_schema, serialization_mapper: dict = None):
    """
    In case field is None, should return None.
    Should deal with a list of fields, as well as a single one
    """
    if field is None:
        return None
    if isinstance(field, ClassReference):
        return _map_class_reference(field, definitions_schema)
    if isinstance(field, list):
        return [
            convert_to_schema(
                f, definitions_schema, serialization_mapper=serialization_mapper
            )
            for f in field
        ]
    custom = field.to_json_schema()
    if custom:
        return custom
    mapper = get_mapper(field.__class__)(field)
    return mapper.to_schema(
        definitions_schema, serialization_mapper=serialization_mapper
    )


def _validated_mapped_value(mapper, key):
    if key in mapper:
        key_mapper = mapper[key]
        if isinstance(key_mapper, (FunctionCall,)):
            logging.error(
                f"{key} is mapped to a function  in serialization mapper- "
                "This is unsupported by code-to-schema conversion. "
                "You will need to manually fix it."
            )
        elif key_mapper is DoNotSerialize or isinstance(key_mapper, Constant):
            return key_mapper
        elif not isinstance(key_mapper, (FunctionCall, str)):
            raise TypeError("mapper must have a FunctionCall or a string")

    return None


def structure_to_schema(structure, definitions_schema, serialization_mapper=None):
    """
    Generate JSON schema from :class:`Structure`
    `See working examples in tests. <https://github.com/loyada/typedpy/tree/master/tests/schema_mapping>`_

    Arguments:
        structure( subclass of :class:`Structure` ):
            the class
        definitions_schema(dict):
            the json schema for all the definitions (typically under "#/definitions" in the schema.
            If it is the first call, just use and empty dict.

    Returns:
        A tuple of 2. The fist is the schema of structure, the second is the schema
        for the referenced definitions.
        the The schema that the code maps to. It also updates

    """
    # json schema draft4 does not support inheritance, so we don't need to worry about that
    props = structure.__dict__
    if not issubclass(structure, Structure):
        raise TypeError("Expected a Structure subclass")
    field_by_name = structure.get_all_fields_by_name()
    required = getattr(structure, "_required", list(field_by_name.keys()))
    additional_props = getattr(structure, ADDITIONAL_PROPERTIES, True)
    mapper = aggregate_serialization_mappers(structure, serialization_mapper) or {}
    if getattr(structure, "_additional_serialization") != getattr(
        Structure, "_additional_serialization"
    ):
        logging.warning(
            "mapping to schema does not support _additional_serialization method. You"
            " will have to edit it manually."
        )
    if (
        len(field_by_name) == 1
        and set(required) == set(field_by_name.keys())
        and additional_props is False
    ):
        _, value = first_in(field_by_name.items())
        return (
            convert_to_schema(value, definitions_schema),
            definitions_schema,
        )
    else:
        fields_schema = OrderedDict([("type", "object")])
        fields_schema["properties"] = {}
        properties = fields_schema["properties"]
        for key, field in field_by_name.items():
            mapped_key = (
                mapper[key]
                if key in mapper and isinstance(mapper[key], (str,))
                else key
            )
            mapped_value = _validated_mapped_value(mapper, key)
            if mapped_value is DoNotSerialize or isinstance(mapped_value, Constant):
                if mapped_key in required:
                    required.pop(required.index(mapped_key))
            else:
                if key in required:
                    required[required.index(key)] = mapped_key
                sub_mapper = mapper.get(f"{key}._mapper", {})
                sub_schema = convert_to_schema(
                    field, definitions_schema, serialization_mapper=sub_mapper
                )
                default_raw = getattr(field, "_default")
                if default_raw is not None:
                    default_val = (
                        default_raw() if callable(default_raw) else default_raw
                    )
                    sub_schema["default"] = default_val
                properties[mapped_key] = sub_schema
        fields_schema.update(
            OrderedDict(
                [
                    ("required", sorted(required)),
                    ("additionalProperties", additional_props),
                ]
            )
        )

    return fields_schema, definitions_schema


@default_factories
def convert_to_field_code(schema, definitions, additional_fields=list):
    """
    In case schema is None, should return None.
    Should deal with a schema that is a dict, as well as one that is a list
    """

    if schema is None:
        return None
    if isinstance(schema, list):
        fields = [convert_to_field_code(s, definitions) for s in schema]
        return f"[{', '.join(fields)}]"
    if "$ref" in schema:
        def_name = schema["$ref"][len("#/definitions/") :]
        return def_name

    type_name_to_field = {
        "object": StructureReference,
        "integer": Integer,
        "number": Number,
        "float": Float,
        "array": Array,
        "string": String,
        "boolean": Boolean,
    }
    multivals = {"allOf": AllOf, "anyOf": AnyOf, "oneOf": OneOf, "not": NotField}
    if any(multival in schema for multival in multivals):
        for (k, the_class) in multivals.items():
            if k in schema:
                cls = the_class
        mapper = MultiFieldMapper

    elif "enum" in schema:
        cls = Enum
        mapper = get_mapper(cls)
    else:
        object_type = schema.get("type", "object")
        if object_type == "object":
            if SCHEMA_PROPETIES in schema:
                cls = StructureReference
            else:
                cls = Map
        else:
            for c in additional_fields:
                custom_mapping = c.from_json_schema(schema)
                if custom_mapping:
                    return custom_mapping
            cls = type_name_to_field[schema.get("type", "object")]
        mapper = get_mapper(cls)
    params_list = mapper.get_paramlist_from_schema(schema, definitions)
    if "default" in schema:
        default_val = schema["default"]
        if isinstance(default_val, (list, dict)):
            default_val = f"lambda: {default_val}"
        params_list.append(("default", default_val))
    params_as_string = ", ".join([f"{name}={val}" for (name, val) in params_list])
    return f"{cls.__name__}({params_as_string})"


@default_factories
def schema_to_struct_code(
    struct_name, schema, definitions_schema, additional_fields=list
):
    """
    Generate code for the main class that maps to the given JSON schema.
    The main struct_name can include references to structures defined in
    definitions_schema, under "#/definitions/".

    Arguments:
        struct_name(str):
            the name of the main :class:`Structure` to be created
        schema(dict):
            the json schema of the main Structure that need to be defined
        definitions_schema(dict):
            schema for definitions of objects that can be referred to in the main schema. If non exist,
            just use an empty dict.
        additional_fields(list):
            additional Types of Fields with custom schema mapping that can appear in the schema. These
            have to implement the class method from_json_schema(), which should return a string of the code
            the Schema is mapping to.
    Returns:
        A string with the code of the class. This can either be executed directly,
        using exec(), or written to a file.
        The "description" property, if exists, is mapped to the docstring of the class.
        If you write to a file, the higher level :func:`write_code_from_schema` is preferable.
        Note: In case schema is None, should return None.
        Deals with a schema that is a dict, as well as one that is a list
    """
    body = [f"class {struct_name}(Structure):"]
    body += (
        [f'    """\n    {schema.get("description")}\n    """\n']
        if "description" in schema
        else []
    )
    body += (
        ["    _additionalProperties = False"]
        if not schema.get("additionalProperties", True)
        else []
    )
    required = (
        schema.get("required", None)
        if schema.get("type", "object") == "object"
        else ["wrapped"]
    )
    the_type = schema.get("type", "object" if "properties" in schema else None)

    if the_type == "object":
        properties = schema.get("properties", {})
        for (name, sch) in properties.items():
            body += [
                f"    {name} = {convert_to_field_code(sch, definitions_schema, additional_fields=additional_fields)}"
            ]
    else:
        body += [
            f"    wrapped = {convert_to_field_code(schema, definitions_schema, additional_fields=additional_fields)}"
        ]

    body += ["", f"    _required = {required}"] if required is not None else []

    return "\n".join(body)


@default_factories
def schema_definitions_to_code(schema, additional_fields=list):
    """
    Generate code for the classes in the definitions that maps to the given JSON schema.
    `See working example in test_schema_to_code.py. <https://github.com/loyada/typedpy/tree/master/tests/test_schema_to_code.py>`_

    Arguments:
        schema(dict):
            the json schema of the various Structures that need to be defined
    Returns:
        A string with the code. This can either be executed directly, using exec(), or written to a file.
        If you write to a file, the higher level :func:`write_code_from_schema` is preferable.
    """
    code = []
    for (name, sch) in schema.items():
        code.append(
            schema_to_struct_code(
                name, sch, schema, additional_fields=additional_fields
            )
        )
    return "\n\n\n".join(code)


@default_factories
def write_code_from_schema(
    schema, definitions_schema, filename, class_name, additional_fields=list
):
    """
    Generate code from schema and write it to a file.

    Example:

    .. code-block:: python

        write_code_from_schema(schema, definitions, "generated_sample.py", "Poo", additional_fields=[CustomField1, CustomField2])


    Arguments:
        schema(dict):
            the json schema for the main structure
        definitions_schema(dict):
            the json schema for all the definitions (typically under "#/definitions" in the schema.
            These can be referred to from the main schema
        filename(str):
            the file name for the output. Typically should be end with ".py".
        class_name(str):
            the main Structure name
        additional_fields(list[Type[Field]]):
            additional field classes with custom mapping
    """
    supporting_classes = schema_definitions_to_code(
        definitions_schema, additional_fields=additional_fields
    )
    structure_code = schema_to_struct_code(
        class_name, schema, definitions_schema, additional_fields=additional_fields
    )
    with open(filename, "w", encoding="utf-8") as fout:
        fout.write("from typedpy import *\n\n\n")
        if definitions_schema:
            fout.write(supporting_classes)
            fout.write("\n\n# ********************\n\n\n")
        fout.write(structure_code)
        fout.write("\n")


class Mapper:
    def __init__(self, value):
        self.value = value


class StructureReferenceMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        body = []
        body += (
            [(ADDITIONAL_PROPERTIES, False)]
            if not schema.get("additionalProperties", True)
            else []
        )
        required = schema.get("required", None)
        body += [("_required", required)] if required is not None else []
        properties = schema.get("properties", {})

        body += [
            (k, convert_to_field_code(v, definitions)) for (k, v) in properties.items()
        ]
        return body

    def to_schema(self, definitions, serialization_mapper):
        schema, _ = structure_to_schema(
            getattr(self.value, "_newclass"), definitions, serialization_mapper
        )
        schema["type"] = "object"
        return schema


class NumberMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        params = {
            "multiplesOf": schema.get("multiplesOf", None),
            "minimum": schema.get("minimum", None),
            "maximum": schema.get("maximum", None),
            "exclusiveMaximum": schema.get("exclusiveMaximum", None),
        }
        return list((k, v) for k, v in params.items() if v is not None)

    def to_schema(self, definitions, serialization_mapper):
        def get_min(value):
            if value.minimum is not None:
                return value.minimum
            if isinstance(value, NonNegative):
                return 0
            if isinstance(value, Positive):
                return 1 if isinstance(value, Integer) else 0.000001
            return None

        def get_max(value):
            if value.maximum is not None:
                return value.maximum
            if isinstance(value, NonPositive):
                return 0
            if isinstance(value, Negative):
                return -1 if isinstance(value, Integer) else -0.000001
            return None

        value = self.value
        params = {
            "type": "number",
            "multiplesOf": value.multiplesOf,
            "minimum": get_min(value),
            "maximum": get_max(value),
            "exclusiveMaximum": value.exclusiveMaximum,
        }
        return {k: v for k, v in params.items() if v is not None}


class MapMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        additional_properties = schema.get(SCHEMA_ADDITIONAL_PROPERTIES, None)
        pattern_properties = schema.get(SCHEMA_PATTERN_PROPERTIES, None)
        property_names = schema.get(SCHEMA_PATTERN_PROPERTIES, {})
        if (
            (pattern_properties and len(pattern_properties) > 1)
            or (pattern_properties and additional_properties)
            or (pattern_properties and property_names)
        ):
            raise NotImplementedError("Conversion for this map is unsupported")
        if not any([additional_properties, property_names, pattern_properties]):
            return []

        key_type = (
            convert_to_field_code({**property_names, "type": "string"}, globals())
            if property_names
            else "String()"
        )
        adjusted_key_type = (
            f"String(pattern='{list(pattern_properties.keys())[0]}')"
            if pattern_properties
            else key_type
        )
        value_type = (
            convert_to_field_code(additional_properties, definitions)
            if additional_properties
            else (
                convert_to_schema(pattern_properties, definitions)
                if additional_properties
                else "Anything"
            )
        )

        items = f"[{adjusted_key_type}, {value_type}]"
        params = {
            "items": items,
            "maxItems": schema.get("maxItems", None),
            "minItems": schema.get("minItems", None),
        }
        return list((k, v) for k, v in params.items() if v is not None)

    def to_schema(self, definitions, serialization_mapper):
        value = self.value
        params = {
            "type": "object",
        }
        if value.items:
            keys, values = value.items
            if not isinstance(keys, String):
                raise TypeError("JSON supports only Strings as keys")

            suffix = ""
            if keys.maxLength or keys.minLength:
                suffix = f"{{{keys.minLength or ''}, {keys.maxLength or ''}}}"
            pattern_props = f"{keys.pattern or ''}{suffix}" or None
            values_schema = convert_to_schema(
                values, definitions, serialization_mapper=serialization_mapper
            )
            if pattern_props:
                params[SCHEMA_PATTERN_PROPERTIES] = values_schema
            elif values_schema:
                params[SCHEMA_ADDITIONAL_PROPERTIES] = values_schema
        params["maxItems"] = value.maxItems
        params["minItems"] = value.minItems
        return {k: v for k, v in params.items() if v is not None}


class IntegerMapper(NumberMapper):
    def to_schema(self, definitions, serialization_mapper):
        params = super().to_schema(definitions, serialization_mapper)
        params.update({"type": "integer"})
        return params


class FloatMapper(NumberMapper):
    def to_schema(self, definitions, serialization_mapper):
        params = super().to_schema(definitions, serialization_mapper)
        params.update({"type": "float"})
        return params


class BooleanMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        return []

    def to_schema(self, definitions, serialization_mapper):  # pylint: disable=R0201
        params = {
            "type": "boolean",
        }
        return params


class StringMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        params = {
            "minLength": schema.get("minLength", None),
            "maxLength": schema.get("maxLength", None),
            "pattern": wrap_val(schema.get("pattern", None)),
        }
        return list((k, v) for k, v in params.items() if v is not None)

    def to_schema(self, definitions, serialization_mapper):
        value = self.value
        params = {
            "type": "string",
            "minLength": value.minLength,
            "maxLength": value.maxLength,
            "pattern": value.pattern,
        }
        return {k: v for k, v in params.items() if v is not None}


class DateStringMapper(Mapper):
    def to_schema(
        self, definitions, serialization_mapper
    ):  # pylint: disable=no-self-use
        params = {
            "type": "string",
            "pattern": r"^([12]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01]))$",
        }
        return {k: v for k, v in params.items() if v is not None}


class ArrayMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        items = schema.get("items", None)
        params = {
            "uniqueItems": schema.get("uniqueItems", None),
            "additionalItems": schema.get("additionalItems", None),
            "items": convert_to_field_code(items, definitions),
        }
        return list((k, v) for k, v in params.items() if v is not None)

    def to_schema(self, definitions, serialization_mapper):
        value = self.value
        if isinstance(value, Tuple):
            params = {
                "type": "array",
                "uniqueItems": value.uniqueItems,
                "additionalItems": False,
                "items": convert_to_schema(
                    value.items, definitions, serialization_mapper
                ),
            }
        elif isinstance(value, Set):
            params = {
                "type": "array",
                "uniqueItems": True,
                "maxItems": value.maxItems,
                "minItems": value.minItems,
                "items": convert_to_schema(
                    value.items, definitions, serialization_mapper
                ),
            }
        else:
            params = {
                "type": "array",
                "uniqueItems": value.uniqueItems,
                "additionalItems": value.additionalItems,
                "maxItems": value.maxItems,
                "minItems": value.minItems,
                "items": convert_to_schema(
                    value.items, definitions, serialization_mapper
                ),
            }
        return {k: v for k, v in params.items() if v is not None}


class EnumMapper(Mapper):
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        params = {
            "values": schema.get("enum", None),
        }
        return list(params.items())

    def to_schema(self, definitions, serialization_mapper):
        def adjust(val) -> Union[str, int, float]:
            if isinstance(val, enum.Enum):
                return (
                    val.value
                    if getattr(self.value, "serialization_by_value", False)
                    else val.name
                )
            if not isinstance(val, (int, str, float)):
                raise TypeError("enum must be an enum, str, or number")
            return val

        values = [adjust(v) for v in self.value.values]
        params = {"enum": values}
        return {k: v for k, v in params.items() if v is not None}


class MultiFieldMapper:
    @staticmethod
    def get_paramlist_from_schema(schema, definitions):
        items = list(schema.values())[0]
        params = {"fields": convert_to_field_code(items, definitions)}
        return list(params.items())


class AllOfMapper(Mapper):
    def to_schema(self, definitions, serialization_mapper):
        return {
            "allOf": convert_to_schema(
                self.value._fields, definitions, serialization_mapper
            )
        }


class OneOfMapper(Mapper):
    def to_schema(self, definitions, serialization_mapper):
        return {
            "oneOf": convert_to_schema(
                self.value._fields, definitions, serialization_mapper
            )
        }


class AnyOfMapper(Mapper):
    def to_schema(self, definitions, serialization_mapper):
        if (
            len(self.value._fields) == 2
            and self.value._fields[1].__class__ == NoneField
        ):
            return convert_to_schema(
                self.value._fields[0], definitions, serialization_mapper
            )
        return {
            "anyOf": convert_to_schema(
                self.value._fields, definitions, serialization_mapper
            )
        }


class NotFieldMapper(Mapper):
    def to_schema(self, definitions, serialization_mapper):
        return {
            "not": convert_to_schema(
                self.value._fields, definitions, serialization_mapper
            )
        }



"""
A type-safe strictly defined structures, compatible with JSON draft 4
but offers significantly more functionality.
"""
from typedpy.structures import (
    Structure,
    Field,
    TypedField,
    ClassReference,
    ImmutableStructure,
    create_typed_field,
    NoneField,
    FinalStructure,
    ImmutableField,
    unique,
    AbstractStructure,
)

from typedpy.structures_reuse import (
    Partial,
    AllFieldsRequired,
    Omit,
    Pick,
    Extend,
)

from typedpy.fields import (
    Number,
    Integer,
    PositiveInt,
    PositiveFloat,
    NonPositiveInt,
    NonPositiveFloat,
    NegativeInt,
    NegativeFloat,
    NonNegativeInt,
    NonNegativeFloat,
    Float,
    Positive,
    Negative,
    NonPositive,
    NonNegative,
    DecimalNumber,
    String,
    SizedString,
    Sized,
    AllOf,
    AnyOf,
    OneOf,
    NotField,
    Boolean,
    Array,
    Set,
    Map,
    Tuple,
    StructureReference,
    Anything,
    SerializableField,
    Function,
    ImmutableMap,
    ImmutableArray,
    ImmutableSet,
    ImmutableFloat,
    ImmutableString,
    ImmutableInteger,
    ImmutableNumber,
    ImmutableDeque,
    Deque,
    ExceptionField,
    Generator,
)

from typedpy.json_schema_mapping import (
    structure_to_schema,
    schema_to_struct_code,
    schema_definitions_to_code,
    write_code_from_schema,
)

from .enum import Enum, EnumString

from .serialization import (
    deserialize_structure,
    serialize,
    serialize_field,
    FunctionCall,
    deserialize_single_field,
    HasTypes,
)

from .serialization_wrappers import (
    Serializer,
    Deserializer,
    deserializer_by_discriminator,
)

from .extfields import (
    DateString,
    DateField,
    DateTime,
    TimeString,
    HostName,
    IPV4,
    JSONString,
    EmailAddress,
)

from .subclass import SubClass

from .errors import (
    standard_readable_error_for_typedpy_exception,
    ErrorInfo,
    get_simplified_error,
)

from .utility import get_list_type, type_is_generic

from .mappers import mappers, Deleted, Constant, DoNotSerialize

from .versioned_mapping import convert_dict, Versioned

from .commons import (
    nested,
    deep_get,
    first_in,
    flatten,
    default_factories,
    InvalidStructureErr,
)

from .keysof import keys_of
from .type_helpers import create_pyi
